import { useState, useEffect } from "react";

const shallowEq = (x, y) => x === y;

const mapValues = (obj, fn) => Object.fromEntries(Object.entries(obj).map(([key, val]) => [key, fn(val)]))

export const createStore = (initial, actions = {}) => {
  const store = {};
  const hooks = {};
  Object.entries(initial).forEach(([key, value]) => {
    let _val = value;
    let _listeners = [];

    store[key] = {
      _get: () => _val,
      _set: newValue => {
        const oldValue = _val;
        _val = newValue;
        _listeners
          .filter(({ equality }) => !equality(oldValue, newValue))
          .forEach(({ callback }) => callback(newValue));
      },
    };

    // name allows linters to pick this up as a hook
    const useValue = (equality = shallowEq) => {
      const [val, callback] = useState(_val);
      useEffect(() => {
        _listeners.push({ callback, equality });
        return () => {
          _listeners = _listeners.filter(ln => ln.callback !== callback);
        }
      }, [equality]);
      return val;
    }
    const hookName = "use" + key.charAt(0).toUpperCase() + key.slice(1);
    hooks[hookName] = useValue;
  });

  const dispatch = mapValues(actions, act => (...args) => Object.entries(act(...args)).forEach(([key, newValue]) => store[key]?._set(newValue)));

  const selector = mapValues(store, obs => obs._get);

  return [hooks, dispatch, selector];
}